//////////////////////////////////
//  thermal camera test program
//////////////////////////////////

#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <poll.h>
#include <pthread.h>

#include "NvEglRenderer.h"
#include "NvUtils.h"
#include "nvbuf_utils.h"

#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <cudaEGL.h>

#include "camera_v4l2.h"
#include "y2yuv.h"

static bool quit = false;

using namespace std;

static void
print_usage(void)
{
    printf("\n\tUsage: camera_v4l2 [OPTIONS]\n\n"
           "\tExample: \n"
           "\t./camera_v4l2_cuda -d /dev/video0 -s 640x480 -f YUYV -n 30 -c\n\n"
           "\tSupported options:\n"
           "\t-d\t\tSet V4l2 video device node\n"
           "\t-s\t\tSet output resolution of video device\n"
           "\t-r\t\tSet renderer frame rate (30 fps by default)\n"
           "\t-n\t\tSave the n-th frame before VIC processing\n"
           "\t-v\t\tEnable verbose message\n"
           "\t-h\t\tPrint this usage\n\n"
           "\tNOTE: It runs infinitely until you terminate it with <ctrl+c>\n");
}

static bool
parse_cmdline(context_t * ctx, int argc, char **argv)
{
    int c;

    if (argc < 2)
    {
        print_usage();
        exit(EXIT_SUCCESS);
    }

    while ((c = getopt(argc, argv, "d:s:f:r:n:cvh")) != -1)
    {
        switch (c)
        {
            case 'd':
                ctx->cam_devname = optarg;
                break;
            case 's':
                if (sscanf(optarg, "%dx%d",
                            &ctx->cam_w, &ctx->cam_h) != 2)
                {
                    print_usage();
                    return false;
                }
                break;
             case 'r':
                ctx->fps = strtol(optarg, NULL, 10);
                break;
            case 'n':
                ctx->save_n_frame = strtol(optarg, NULL, 10);
                break;
             case 'v':
                ctx->enable_verbose = true;
                break;
            case 'h':
                print_usage();
                exit(EXIT_SUCCESS);
                break;
            default:
                print_usage();
                return false;
        }
    }

    return true;
}

static void
set_defaults(context_t * ctx)
{
    memset(ctx, 0, sizeof(context_t));

    ctx->cam_devname = "/dev/video0";
    ctx->cam_fd = -1;
    ctx->cam_pixfmt = V4L2_PIX_FMT_GREY;
    ctx->cam_w = 640;
    ctx->cam_h = 512;
    ctx->frame = 0;
    ctx->save_n_frame = 0;

    ctx->g_buff = NULL;
    ctx->renderer = NULL;
    ctx->got_error = false;
    ctx->fps = 30;
    ctx->o_buff.fd = -1;

    ctx->output_plane_buf_queue = new queue < nv_buffer * >;
    pthread_mutex_init(&ctx->queue_lock, NULL);
    pthread_cond_init(&ctx->queue_cond, NULL);

    ctx->enable_cuda = true;
    ctx->egl_image = NULL;
    ctx->egl_display = EGL_NO_DISPLAY;
    ctx->cam_image = NULL;
    ctx->cuContext = 0;

    ctx->enable_verbose = false;
    sprintf(ctx->cam_file, "camera.raw");

}

static bool
save_frame_to_file(char * filename, unsigned char * startp, int numbytes)
{
    int file;

    file = open(filename, O_CREAT | O_WRONLY | O_APPEND | O_TRUNC,
            S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

    if (-1 == file)
        ERROR_RETURN("Failed to open file for frame saving %s ", filename);

    if (-1 == write(file, startp, numbytes))
    {
        close(file);
        ERROR_RETURN("Failed to write frame into file");
    }

    close(file);

    return true;
}

static bool
camera_initialize(context_t * ctx)
{
    struct v4l2_format fmt;

    // Open camera device
    ctx->cam_fd = open(ctx->cam_devname, O_RDWR);
    if (ctx->cam_fd == -1)
        ERROR_RETURN("Failed to open camera device %s: %s (%d)",
                ctx->cam_devname, strerror(errno), errno);

    // Set camera output format
    memset(&fmt, 0, sizeof(fmt));
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width = ctx->cam_w;
    fmt.fmt.pix.height = ctx->cam_h;
    fmt.fmt.pix.pixelformat = ctx->cam_pixfmt;
    fmt.fmt.pix.field = V4L2_FIELD_NONE;
    if (ioctl(ctx->cam_fd, VIDIOC_S_FMT, &fmt) < 0)
        ERROR_RETURN("Failed to set camera output format: %s (%d)",
                strerror(errno), errno);

    // Get the real format in case the desired is not supported.
    memset(&fmt, 0, sizeof fmt);
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(ctx->cam_fd, VIDIOC_G_FMT, &fmt) < 0)
        ERROR_RETURN("Failed to get camera output format: %s (%d)",
                strerror(errno), errno);
    if (fmt.fmt.pix.width != ctx->cam_w ||
            fmt.fmt.pix.height != ctx->cam_h )
    {
        WARN("The desired frame size is not supported");
        ctx->cam_w = fmt.fmt.pix.width;
        ctx->cam_h = fmt.fmt.pix.height;
    }

    INFO("Camera ouput format: (%d x %d)  stride: %d, imagesize: %d, format: %4x",
            fmt.fmt.pix.width,
            fmt.fmt.pix.height,
            fmt.fmt.pix.bytesperline,
            fmt.fmt.pix.sizeimage,
            fmt.fmt.pix.pixelformat);

    return true;
}

static bool
display_initialize(context_t * ctx)
{
 
    if (ctx->enable_cuda)
    {
		
        // Get default EGL display
        ctx->egl_display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
        if (ctx->egl_display == EGL_NO_DISPLAY)
            ERROR_RETURN("Failed to get EGL display connection");

        // Init EGL display connection
        if (!eglInitialize(ctx->egl_display, NULL, NULL))
            ERROR_RETURN("Failed to initialize EGL display connection");
    }

    // Create EGL renderer
    ctx->renderer = NvEglRenderer::createEglRenderer("renderer0",
           ctx->cam_w, ctx->cam_h, 1240, 100);
    if (!ctx->renderer)
       ERROR_RETURN("Failed to create EGL renderer");
    ctx->renderer->setFPS(ctx->fps);

    return true;
}

static bool
init_components(context_t * ctx)
{
    if (!camera_initialize(ctx))
        ERROR_RETURN("Failed to initialize camera device");

    if (!display_initialize(ctx))
        ERROR_RETURN("Failed to initialize display");

    INFO("Initialize v4l2 components successfully");
    return true;
}

static bool
request_camera_buff(context_t *ctx)
{
    // Request camera v4l2 buffer
    struct v4l2_requestbuffers rb;
    memset(&rb, 0, sizeof(rb));
    rb.count = V4L2_BUFFERS_NUM;
    rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    rb.memory = V4L2_MEMORY_DMABUF;
    if (ioctl(ctx->cam_fd, VIDIOC_REQBUFS, &rb) < 0)
        ERROR_RETURN("Failed to request v4l2 buffers: %s (%d)",
                strerror(errno), errno);
    if (rb.count != V4L2_BUFFERS_NUM)
        ERROR_RETURN("V4l2 buffer number is not as desired");
		
    for (unsigned int index = 0; index < V4L2_BUFFERS_NUM; index++)
    {
        struct v4l2_buffer buf;

        // Query camera v4l2 buf length
        memset(&buf, 0, sizeof buf);
        buf.index = index;
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_DMABUF;

        if (ioctl(ctx->cam_fd, VIDIOC_QUERYBUF, &buf) < 0)
            ERROR_RETURN("Failed to query buff: %s (%d)",
                    strerror(errno), errno);

        // Enqueue empty v4l2 buff into camera capture plane
	buf.m.fd = (unsigned long)ctx->g_buff[index].fd;
	if (buf.length != ctx->g_buff[index].size)
	{
	    WARN("Camera v4l2 buffer is unexpected size cam=%d buf = %d", buf.length, ctx->g_buff[index].size);
	    ctx->g_buff[index].size = buf.length;
	}
        INFO("buffer index %d and file descriptor %x", index, buf.m.fd)

        if (ioctl(ctx->cam_fd, VIDIOC_QBUF, &buf) < 0)
            ERROR_RETURN("Failed to enqueue buffers: %s (%d)",
                    strerror(errno), errno);
    }

    return true;
}

static bool
prepare_buffers(context_t * ctx)
{
    int fd;
    NvBufferParams params={0};
    // Allocate global buffer context
    ctx->g_buff = (nv_buffer *)malloc(V4L2_BUFFERS_NUM * sizeof(nv_buffer));
    if (ctx->g_buff == NULL)
        ERROR_RETURN("Failed to allocate global buffer context");
	
    // Create buffers to share between camera and host
    for (unsigned int index=0; index < V4L2_BUFFERS_NUM; index++)
    {
	if (-1 == NvBufferCreate(&fd, ctx->cam_w, ctx->cam_h, NvBufferLayout_Pitch, NvBufferColorFormat_NV12))
		ERROR_RETURN("Failed to create NvBuffer");
	ctx->g_buff[index].fd = fd;
        ctx->g_buff[index].index = index;
		
	if (-1 == NvBufferGetParams(fd, &params))
		ERROR_RETURN("Failed to get NvBuffer parameters");
		
	// since this is a grey scale image, we only care about the first plane. 
	// Two plane format NV12 is used rather than a single interleaved plane
        // so that plane 0 is sized correctly.  The NvBuffer class has no grey 
        // or Bayer option.
	ctx->g_buff[index].size = params.height[0] * params.pitch[0];
	ctx->g_buff[index].start = (unsigned char *) mmap( NULL, ctx->g_buff[index].size, 
	               PROT_READ | PROT_WRITE, MAP_SHARED, ctx->g_buff[index].fd, 0);
    }

    if (!request_camera_buff(ctx))
        ERROR_RETURN("Failed to set up camera buff");
		
    // Create output buffer for rendering
    if (-1 == NvBufferCreate(&fd, ctx->cam_w, ctx->cam_h,
	                NvBufferLayout_Pitch, NvBufferColorFormat_YUV420))
	    ERROR_RETURN("Failed to create Output buffer");
    ctx->o_buff.fd = fd;
	
    if (-1 == NvBufferGetParams(fd, &params))
	ERROR_RETURN("Failed to get NvBuffer parameters for output buffer");
    ctx->o_buff.totalsize = 0;
    for (int ii=0; ii<params.num_planes; ++ii)
    {
	ctx->o_buff.pitch[ii] = params.pitch[ii];
	ctx->o_buff.size[ii] = params.height[ii] * params.pitch[ii];
	ctx->o_buff.totalsize += ctx->o_buff.size[ii];
        INFO("Buffer plane %d width = %d, height =%d, pitch = %d", ii, params.width[ii], params.height[ii], params.pitch[ii]);
    }
    ctx->o_buff.start = (unsigned char *) mmap( NULL, ctx->o_buff.totalsize, 
	               PROT_READ | PROT_WRITE, MAP_SHARED, ctx->o_buff.fd, 0);
	
    INFO("Succeed in preparing stream buffers");
    return true;
}

static bool
start_stream(context_t * ctx)
{
    enum v4l2_buf_type type;

    // Start v4l2 streaming
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(ctx->cam_fd, VIDIOC_STREAMON, &type) < 0)
        ERROR_RETURN("Failed to start streaming: %s (%d)",
                strerror(errno), errno);

    usleep(200);

    INFO("Camera video streaming on ...");
    return true;
}

static void
abort(context_t *ctx)
{
    ctx->got_error = true;
}

void * consumer_execute(void * arg)
{
    context_t *ctx = (context_t *) arg;
    nv_buffer * cam_g_buff;
    struct v4l2_buffer buf;
    bool done = false;
    int framecnt = 0;
    
    CUresult status;
    CUeglFrame rawFrame;
    CUeglFrame yuvFrame;
    CUgraphicsResource pResourceRaw;
    CUgraphicsResource pResourceYuv;
    const char * errorString;
	
     // initialize cuda context
    status = cuInit(0);
    if (status != CUDA_SUCCESS)
    {
       cuGetErrorString(status, &errorString);
       fprintf(stderr, "Unable to initialize CUDA driver api %s\n", errorString);
       done = true;
    }
    int devcnt;
    status = cuDeviceGetCount(&devcnt);
    if (status != CUDA_SUCCESS)
    {
       cuGetErrorString(status, &errorString);
       fprintf(stderr,"Unable to get number of gpu devices %s\n", errorString);
       done = true;
    }
    if (devcnt == 0)
    {
	fprintf(stderr,"No cuda capable gpu devices\n");			
    }
    CUdevice devhandle;
    status = cuDeviceGet(&devhandle, 0);
    if (status != CUDA_SUCCESS)
    {
       cuGetErrorString(status, &errorString);
       fprintf(stderr,"Unable to get handle to gpu device %s\n", errorString);
       done = true;
    }
    status = cuCtxCreate(&ctx->cuContext, 0, devhandle);
    if (status != CUDA_SUCCESS)
    {
       cuGetErrorString(status, &errorString);
       fprintf(stderr,"Unable to initialize CUDA driver API %s\n", errorString);
       done = true;
    }

    while (!done)
    {

       // Fetch nv_buffer
       pthread_mutex_lock(&ctx->queue_lock);
       while (ctx->output_plane_buf_queue->empty())
       {
           pthread_cond_wait(&ctx->queue_cond, &ctx->queue_lock);
       }
       cam_g_buff = ctx->output_plane_buf_queue->front();
       ctx->output_plane_buf_queue->pop();
       pthread_mutex_unlock(&ctx->queue_lock);

       //fprintf(stdout, "Received signal that buffer is ready\n");
       framecnt++;

       // Got EOS signal and return
       if (cam_g_buff->fd == 0)
       {
           fprintf(stderr, "File descriptor is zero, stop processing\n");
           done = true;
           continue;
       }

       // Create EGLImage from camera buffer fd
       ctx->cam_image = NvEGLImageFromFd(ctx->egl_display, cam_g_buff->fd);
       if (ctx->cam_image == NULL)
       {
           fprintf(stderr, "Failed to map camera buf fd (0x%X) to EGLImage",
                    cam_g_buff->fd);
           done = true;
           continue;
       }
       status = cuGraphicsEGLRegisterImage(&pResourceRaw, ctx->cam_image, CU_GRAPHICS_MAP_RESOURCE_FLAGS_READ_ONLY);
       if (status != CUDA_SUCCESS)
       {
	    cuGetErrorString(status, &errorString);
	    fprintf(stderr, "Failure to register raw image %s\n", errorString);
	    done = true;
            continue;
       }
       status = cuGraphicsResourceGetMappedEglFrame(&rawFrame, pResourceRaw, 0, 0);	
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
	    fprintf(stderr, "Failure to get mapped raw image %s\n", errorString);
	    done = true;
            continue;
       }
       if (ctx->frame == 1)
       {
           INFO("Camera EGL Frame w %d h %d p %d planes %d", rawFrame.width, rawFrame.height, rawFrame.pitch, rawFrame.planeCount);
           if (rawFrame.frameType == CU_EGL_FRAME_TYPE_ARRAY)
           {
              INFO("Array type frame");
           }
           else if (rawFrame.frameType == CU_EGL_FRAME_TYPE_PITCH)
           {
               INFO("Pitch type frame");
           }
           else
           {
               INFO("Unknown type frame %d", rawFrame.frameType);
           }
       }

       // Create EGLImage from output buffer fd
       ctx->egl_image = NvEGLImageFromFd(ctx->egl_display, ctx->o_buff.fd);
       if (ctx->egl_image == NULL)
       {
           fprintf(stderr, "Failed to map output buf fd (0x%X) to EGLImage",
                    ctx->o_buff.fd);
           done = true;
           continue;
       }
       status = cuGraphicsEGLRegisterImage(&pResourceYuv, ctx->egl_image, CU_GRAPHICS_MAP_RESOURCE_FLAGS_WRITE_DISCARD);
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
	    fprintf(stderr, "Failure to register color image %s\n", errorString);
	    done = true;
            continue;
       }
       status = cuGraphicsResourceGetMappedEglFrame(&yuvFrame, pResourceYuv, 0, 0);	
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
	    fprintf(stderr, "Failure to get mapped color image %s\n", errorString);
	    done = true;
            continue;
       }
       if (ctx->frame == 1)
       {
          INFO("Color EGL Frame w %d h %d p %d planes %d", yuvFrame.width, yuvFrame.height, yuvFrame.pitch, yuvFrame.planeCount);
          if (yuvFrame.frameType == CU_EGL_FRAME_TYPE_ARRAY)
          {
              INFO("Array type frame");
          }
          else if (yuvFrame.frameType == CU_EGL_FRAME_TYPE_PITCH)
          {
              INFO("Pitch type frame");
          }
          else
          {
              INFO("Unknown type frame %d", yuvFrame.frameType);
          }
       }

       status = cuCtxSynchronize();
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
	    fprintf(stderr, "Synchronization failure %s\n", errorString);
	    done = true;
            continue;
       }
					
       //Convert from Y to YUV
       gpuConvertY2YUV( (CUdeviceptr)rawFrame.frame.pPitch[0], (CUdeviceptr)yuvFrame.frame.pPitch[0], (CUdeviceptr)yuvFrame.frame.pPitch[1], (CUdeviceptr)yuvFrame.frame.pPitch[2],
	                   rawFrame.width, rawFrame.height, rawFrame.pitch, ctx->o_buff.pitch[0], ctx->o_buff.pitch[1]);

       status = cuCtxSynchronize();
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
	    fprintf(stderr, "Synchronization failure after memory copy %s\n", errorString);
	    done = true;
            continue;
       }

       status = cuGraphicsUnregisterResource(pResourceRaw);
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
            fprintf(stderr, "Failure to unregister raw image %s\n", errorString);
	    done = true;
            continue;
      }
       status = cuGraphicsUnregisterResource(pResourceYuv);
       if (status != CUDA_SUCCESS)
       {
            cuGetErrorString(status, &errorString);
            fprintf(stderr, "Failure to unregister color image %s\n", errorString);
	    done = true;
            continue;
       }
		
       // Destroy EGL Images
       NvDestroyEGLImage(ctx->egl_display, ctx->egl_image);
       ctx->egl_image = NULL;
       NvDestroyEGLImage(ctx->egl_display, ctx->cam_image);
       ctx->cam_image = NULL;

 
       if ((ctx->save_n_frame > 0) && (framecnt == ctx->save_n_frame+5))
       {
            char fname[] = {"coloroutput.raw"};
            save_frame_to_file(fname, ctx->o_buff.start, ctx->o_buff.totalsize);
            fprintf(stdout, "writing result to output file %d bytes\n", ctx->o_buff.totalsize);
       }
       // render the processed output buffer
       ctx->renderer->render(ctx->o_buff.fd);
    }

    status = cuCtxDestroy(ctx->cuContext);
    if (status != CUDA_SUCCESS)
    {
	fprintf(stderr, "Unable to destroy CUDA context\n");
    }


}

static void
signal_handle(int signum)
{
    printf("Quit due to exit command from user!\n");
    quit = true;
}



static bool
start_capture(context_t * ctx)
{
    struct sigaction sig_action;
    struct pollfd fds[1];
    pthread_t consumer;
    int pollrtn;

    // Ensure a clean shutdown if user types <ctrl+c>
    sig_action.sa_handler = signal_handle;
    sigemptyset(&sig_action.sa_mask);
    sig_action.sa_flags = 0;
    sigaction(SIGINT, &sig_action, NULL);
	
    //Start consumer thread
     pthread_create(&consumer, NULL, consumer_execute, ctx);

    // Enable render profiling information
    ctx->renderer->enableProfiling();

    fds[0].fd = ctx->cam_fd;
    fds[0].events = POLLIN;
    while (!ctx->got_error && !quit)
    {
        pollrtn = poll(fds,1,5000);   // block until buffer available
        if (pollrtn > 0)
        {
          if (fds[0].revents & POLLIN) 
          {
            struct v4l2_buffer v4l2_buf;

            // Dequeue camera buff
            memset(&v4l2_buf, 0, sizeof(v4l2_buf));
            v4l2_buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            v4l2_buf.memory = V4L2_MEMORY_DMABUF;
            if (ioctl(ctx->cam_fd, VIDIOC_DQBUF, &v4l2_buf) < 0)
                ERROR_RETURN("Failed to dequeue camera buff: %s (%d)",
                        strerror(errno), errno);

            ctx->frame++;

            if ((ctx->save_n_frame > 0) && (ctx->frame == ctx->save_n_frame))
            {
                save_frame_to_file(ctx->cam_file, ctx->g_buff[v4l2_buf.index].start, ctx->g_buff[v4l2_buf.index].size);
                fprintf(stdout, "Writing source frame to file\n");
            }

            // Push nv_buffer into output queue for conversion and display
            pthread_mutex_lock(&ctx->queue_lock);
            ctx->output_plane_buf_queue->push(&ctx->g_buff[v4l2_buf.index]);
            pthread_cond_broadcast(&ctx->queue_cond);
            pthread_mutex_unlock(&ctx->queue_lock);

            // enqueue the buffer for camera capture
            // We should wait until the consumer is done with the buffer,
            // but the processing takes much less than 1 frame time.
            if (ioctl(ctx->cam_fd, VIDIOC_QBUF, &v4l2_buf))
               ERROR_RETURN("Failed to queue camera buffers: %s (%d)", strerror(errno), errno);
          }
        }
        else
        {
           // timeout or error
           if (pollrtn == 0)
           {
               fprintf(stderr, "camera timeout\n");
           }
           else
           {
               fprintf(stderr, "camera error %d \n", fds[0].revents);
           }
           quit = true;
        }
    }


    if (quit)
    {
        // Signal EOS to the consumer thread
        ctx->g_buff[0].fd = 0;

        pthread_mutex_lock(&ctx->queue_lock);
        ctx->output_plane_buf_queue->push(&ctx->g_buff[0]);
        pthread_cond_broadcast(&ctx->queue_cond);
        pthread_mutex_unlock(&ctx->queue_lock);
    }

    // Wait for consumer thread to stop
    if (!ctx->got_error)
    {
        pthread_join(consumer, NULL);
    }

    // Print profiling information when streaming stops.
    ctx->renderer->printProfilingStats();

    return true;
}

static bool
stop_stream(context_t * ctx)
{
    enum v4l2_buf_type type;

    // Stop v4l2 streaming
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(ctx->cam_fd, VIDIOC_STREAMOFF, &type))
        ERROR_RETURN("Failed to stop streaming: %s (%d)",
                strerror(errno), errno);

    INFO("Camera video streaming off ...");
    return true;
}

int
main(int argc, char *argv[])
{
    context_t ctx;
    int error = 0;

    set_defaults(&ctx);

    CHECK_ERROR(parse_cmdline(&ctx, argc, argv), cleanup,
            "Invalid options specified");

    CHECK_ERROR(init_components(&ctx), cleanup,
            "Failed to initialize v4l2 components");

    CHECK_ERROR(prepare_buffers(&ctx), cleanup,
            "Failed to prepare v4l2 buffs");

    CHECK_ERROR(start_stream(&ctx), cleanup,
            "Failed to start streaming");

    CHECK_ERROR(start_capture(&ctx), cleanup,
            "Failed to start capturing")

    CHECK_ERROR(stop_stream(&ctx), cleanup,
            "Failed to stop streaming");

cleanup:
    if (ctx.cam_fd > 0)
        close(ctx.cam_fd);

    if (ctx.renderer != NULL)
        delete ctx.renderer;
  

    if (ctx.egl_display && !eglTerminate(ctx.egl_display))
        fprintf(stderr,"Failed to terminate EGL display connection\n");

    if (ctx.o_buff.fd)
    {
	NvBufferDestroy(ctx.o_buff.fd);
    }

    if (ctx.g_buff != NULL)
    {
        free(ctx.g_buff);
    }
	
    while(!ctx.output_plane_buf_queue->empty())
    {
        ctx.output_plane_buf_queue->pop();
    }
    delete ctx.output_plane_buf_queue;

    if (error)
        fprintf(stderr,"App run failed\n");
    else
        fprintf(stdout, "App run was successful\n");

    return -error;
}
